"""
blisp   -  The best lisp implementation in the world!
by Brendan O'Connor

This file, all.py, is the big mish-mash of stuff that hasn't been
modularized yet.  It's what you run to start the interpreter.

"""
from __future__ import generators
from copy import copy, deepcopy
from Parser import ParseString, QuotedLispList
from Objects import LispList

class LispException(Exception): pass

#################################################################

def EvalExpr(expr, ns):
    """That's expression, namespace, where ns is a table of string-to-string
        mappings.
    expr can be a string, a LispList
    If expr.isQuoted, that is it represents '(1 2) then '(1 2) is returned.
    If it's not quoted, then it gets recursively evaluated, the first elt getting
    called.  And namespace substitution happens, too.

    If a LispList is returned, it is ALWAYS isQuoted.


    THIS FUNCTION IS NOT COPY SAFE needs a writable copy of expr!!!
        But ns is read-only.
    """
    
    print "EvalExpr begin: %s, type %s" %(expr, type(expr))
    if isinstance(expr, str):
        if expr in ns:  return ns[expr]
        else:           return expr

    if hasattr(expr, 'isQuoted') and expr.isQuoted:
        return expr #Don't eval it if quoted!

    if expr[0] in builtin_macros: #We don't want to do any recursive evaluation!
        return builtin_macros[expr[0]](expr[1:])

    if   isinstance(expr, LispList): startfrom = 1
    else: raise "Error in EvalExpr: expr is unhandled type %s" %type(expr)

    Substitute(expr, ns, startfrom)
    print "EvalExpr: After substitution: %s" %expr

    for i in range(startfrom, len(expr)):
        expr[i] = EvalExpr(expr[i], ns)

    if isinstance(expr, LispList):
        expr = CallFunction(expr[0], expr[1:])
    if isinstance(expr, LispList): expr.isQuoted = 1
    return expr


def Substitute(lispList, ns, startfrom=1):
    "Does namespace substitution on lispList, in-place."
    for i in range(startfrom, len(lispList)):  #We iterate over indexes not lispList because if you do that, you can't substitute strings because the list iterator makes copies of them because they're immutable...
        if isinstance(lispList[i], str):
            print 'thinking about substitutions for %s...' %lispList[i]
            if lispList[i] in ns:
                print '\tsubstituting %s for %s!' %(ns[lispList[i]], lispList[i])
                lispList[i] = ns[lispList[i]]
            else:
                print '\tnot subbing.'
        else:
            assert isinstance(lispList[i], LispList)

def CallFunction(fnName, args):
    """Does function-namespaces lookup and then calls the correct function.
        It only exists because the global non-modularized mess that is the
        current namespace system.
    It's expected that the function's arguments have already been resolved.
    args can be any plain ol' sequence
    """
    #Get fn to be a function object that takes one argument,
    #    the lisp function's args.
    if   fnName in builtin_functions:
        fn = builtin_functions[fnName]
    elif fnName in defined_functions:
        fn = defined_functions[fnName].Call
    else:
        raise LispException, "Unknown function name '%s'!" %fnName
    return fn(args)


class LispFunction(object):
    def __init__(self, params, evals_to):
        self.params   = params
        self.evals_to = evals_to
        print "New LispFunction created."
        print "  params=%(params)s, evals_to=%(evals_to)s"  %vars()

    def checkargs(self, args):
        "Helper for Call()"
        if len(args) != len(self.params):
            error_msg = "Call died!, I, %s, got %d args but want %d!" \
                        % (self, len(args), len(self.params))
            raise LispException, error_msg
    def make_namespace(self, args):
        """Helper for Call()
        Example:  If params=['a','b'] and args=[1, 2],
            then we return { 'a':1, 'b':2 }
        """
        items = zip(self.params, args)
        return dict(items)

    def Call(self, args):
        "args have already been recursively evaluated!"
        self.checkargs(args)
        print "Call() begin: args=%s, %s" %(args, type(args))
        thisCallsNamespace = self.make_namespace(args)

##        SubstituteRecurs(self.evals_to, ns)
##        print "%(self)s.Call(): After substitution: %(args)s" % vars()
        print "Call(): created namespace=%s" %thisCallsNamespace
        eval_result = EvalExpr(deepcopy(self.evals_to), thisCallsNamespace)

        if isinstance(eval_result, LispList): eval_result.isQuoted = 1
        print "Call(): returning %s %s" %(eval_result, type(eval_result))
        return eval_result

##############################
#  The builtin functions... The architecture here needs an overhaul.
#   I'm thinking, they should be subclassed from LispFunction.
#
#  You have to call them AFTER recursively evalling what will be their args..
#  Except for defun, or I guess macros in general when there's more of those.
#
# Note: lispargs.isQuoted is IRRELEVANT.. you want to call these AFTER
# quote-dependent processing.
#   Heck, lispargs could be a normal list!  The elements inside of it, however,
#   are atoms and LispLists.

def LispEval(lispargs):
    """\
    Lisp usage:     (eval '(+ 1 2))
    Python usage:   LispEval( [LispList(['+', '1', '2'], isQuoted=1)] )
    
    """
    assert len(lispargs) == 1
    eval_this = lispargs[0]
    assert isinstance(eval_this, LispList) and len(eval_this) > 0, type(eval_this)
    assert eval_this.isQuoted

    eval_this_copy = deepcopy(eval_this)
    eval_this_copy.isQuoted = 0
    return EvalExpr(eval_this_copy, {})
                 
##    if   funcname in builtins:
##        return apply(builtins[funcname], lispargs )
##    elif funcname in defined_functions:
##        caller = defined_functions[funcname]
##        return caller.Call(lispargs)
##    else:
##        print "LispEval doesn't know what's going on with %s" %lispargs
##        return -1

def LispDefun(lispargs):
    global defined_functions #xxx bad hack

    assert len(lispargs) == 3
    name = lispargs[0]; assert isinstance(name, str)
    params = lispargs[1]; assert isinstance(params, LispList)
    evals_to = lispargs[2]
    if name in defined_functions:
        raise LispExeption, \
                "defun error: Function with name '%s' already exists" % name

    defined_functions[name] = LispFunction(params, evals_to)
    return 0

    
def LispAdd(lispargs):
    assert len(lispargs) == 2
    for e in lispargs: assert isinstance(e, str), e
    numeric_sum = int(lispargs[0]) + int(lispargs[1])
    return str(numeric_sum)

## NOTE: Since there are no cons cells in blisp, we only support lisp-lists.
##       This means there are no dotted pairs and related things I don't know about
##       that lisp is supposed to have.
def LispCons(lispargs):
    assert len(lispargs) == 2 
    ret = LispList( [lispargs[0]] + lispargs[1] )
    return ret

def LispCdr(lispargs):
    assert len(lispargs) == 1
    return LispList( lispargs[0][1:] )
    
def LispCar(lispargs):
    assert len(lispargs) == 1
    ret = lispargs[0][0]
    return ret

    
builtin_functions = {
    'eval'         : LispEval,
    'car'          : LispCar,
    'cdr'          : LispCdr,
    'cons'         : LispCons,
    '+'            : LispAdd
}

#Macros don't get their arguments evaluated first.  Functions do.
builtin_macros = {'defun'   : LispDefun
                 }

defined_functions = {} #Maps name -> LispFunction

print "all.py: All builtin functions:\n", builtin_functions
print "all.py: All builtin macros:\n", builtin_macros

def ProcessString(s):
    if QuotedLispList(s):
        return ParseString(s)
    #We've gotta hack it slightly so (eval ..) or rather, LispEval can handle it
    eval_lisparg  = ParseString( "'%s" %s )
    print "Evalling: ",eval_lisparg
    eval_lispargs = [eval_lisparg] #Normal list is fine for lispargs to a builtin    
    return LispEval( eval_lispargs )


def IILoop():
    "Interactive Interpreter loop"
    print """
This is blisp.
By Brendan O'Connor.  Positively the best LISP ever!
CTRL-C or 'q' to quit."""
    while 1:
        s = raw_input("blisp :-)  ")
        if s == "": continue
        if s[0] == ";": continue
        if s == 'q': return

        print ProcessString(s)

if __name__=='__main__':
    
    IILoop()                      
